@@ -246,7 +246,7 @@ static int data_recv(uint8_t *resp, int length, void *data)
 		tag->current_block = 0;
 
 		DBG("Done reading %zd bytes at %p", data_length, nfc_data);
-		records = near_ndef_parse(nfc_data, data_length);
+		records = near_ndef_parse_msg(nfc_data, data_length, NULL);
 		near_tag_add_records(tag->tag, records, tag->cb, 0);
 
 		g_free(tag);
@@ -288,7 +288,7 @@ static int data_read(struct type3_tag *tag)
 	tag->current_block = 0;
 
 	prepare_read_block(DATA_BLOCK_START + tag->current_block,
-							tag->IDm, &cmd );
+							tag->IDm, &cmd);
 
 	adapter_idx = near_tag_get_adapter_idx(tag->tag);
 
@@ -503,8 +503,7 @@ static int receive_system_code(uint8_t *resp, int length, void *data)
 					cookie, NULL);
 		break;
 
-	case FELICA_LITE_S_IC_TYPE:
-	case FELICA_PLUG_IC_TYPE:
+	default:
 		/* CMD POLL */
 		cmd.cmd	 = CMD_POLL;	/* POLL command */
 		cmd.data[0] = 0x12;     /* System code (NFC SC) */
